"""Top level module for handling API requests"""
import base64
import getpass
import json
import pprint
from _pickle import UnpicklingError
from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import Optional
from typing import Union

import numpy.typing as npt
import requests
from numpy import load
from requests.exceptions import JSONDecodeError
from werkzeug.datastructures import FileStorage
from werkzeug.exceptions import HTTPException, BadRequest, Unauthorized, InternalServerError, Forbidden

from .authorisation import Authorisation
from .plotting import plot

CHASSIS_NAME = 'virtual'
DEFAULT_QUEUE_LENGTH = 10


class Client:
    """Client class for handling API requests"""
    auth: Authorisation = Authorisation(url=None)

    def __init__(self, url=None):
        self.max_queue_length = 50
        if url is not None:
            self.auth.update_urls(url)

    def initialise(self, token_path: Union[Path, str] = None):
        """Create an authenticated session"""
        if token_path is None:
            credentials = self._get_credentials()
            self.auth.authenticate(credentials=credentials)
        else:
            self.auth.authenticate(refresh_token_path=token_path)

    def gen_token(self, token_path: Union[Path, str], credentials: dict = None):
        """Generate an authentication token at a given path"""
        if credentials is None:
            credentials = self._get_credentials()
        self.auth.gen_refresh_token(credentials, token_path)

    def run_emulator(self, qasm_path: Union[Path, str], target_system: str,
                     num_shots: int = 1, noise_percentage=None, allow_transpilation=False):
        """Request for a qasm input to be run on an emulator node"""
        run_result_response = self._run_emulator_request(qasm_path, target_system,
                                                         num_shots, noise_percentage, allow_transpilation)
        self._raise_if_http_error(run_result_response)
        return self._format_emulator_result(run_result_response.json())

    def get_emulator_target_systems(self):
        """Get a list of valid target systems"""
        response = self._run_get_emulator_target_system_request()
        self._raise_if_http_error(response)
        if response.ok:
            return response.json().get('configs')

    def get_emulator_coupling(self, target_system):
        """Get a list of coupled qubits for a given target system"""
        response = self._run_get_emulator_coupling_map_request(target_system)
        self._raise_if_http_error(response)
        if response.ok:
            return response.json().get('coupling')

    def get_experiments(self, start_index: int = 0, end_index: int = DEFAULT_QUEUE_LENGTH):
        """Get previous experiments. Index represents how far back to look."""
        assert end_index > start_index, 'end index must be > start index'
        assert (end_index - start_index) <= self.max_queue_length, 'Please specify a range of experiments less than 50'
        return self._get_experiments(start_index=start_index, end_index=end_index)

    @staticmethod
    def get_experiment(uid: str):
        """Generate an experiment object using an existing experiment ID"""
        return Experiment(uid=uid)

    def create_experiment(self, upload: Union[Path, str]):
        """Create a new experiment from either a qasm string or a path pointing to a string"""
        self._check_file_extension(upload)
        return Experiment(circuit_filepath=upload)

    def _get_hardware_request(self):
        """Make call to get hardware info"""

        @self.auth.request_handler
        def wrapped_get_hardware(url: str, headers: dict):
            """Get the version of the resource API"""
            response = requests.get(f"{url}/api/v1/hardware",
                                    headers=headers)
            return response

        return wrapped_get_hardware()

    def get_hardware(self):
        """Get information on the available hardware backend"""
        return self._get_hardware_request().json()

    def _get_credentials(self):
        username = self._get_username()
        password = self._get_password()
        print('\nAuthenticating with authorisation server')
        return {'email': username, 'password': password}

    @staticmethod
    def _get_username():
        return input('Input your email address\n')

    @staticmethod
    def _get_password():
        return getpass.getpass('Input your password\n')

    @staticmethod
    def _raise_if_http_error(response):
        """Check if the request failed and if so raise the error"""
        if not response.ok:
            print(response.status_code)
            if response.status_code == 400:
                raise BadRequest(response=response, description=response.text)
            if response.status_code == 401:
                raise Unauthorized(response=response, description=response.text)
            if response.status_code == 403:
                raise Forbidden(response=response, description=response.text)
            if response.status_code == 500:
                raise InternalServerError(response=response, description=response.text)
            raise HTTPException(response=response, description=response.text)
        return False

    def _get_version_request(self):
        """Get version api call response"""

        @self.auth.request_handler
        def wrapped_get_version(url: str, headers: dict):
            """Get the version of the resource API"""
            response = requests.get(f"{url}/version",
                                    headers=headers)
            return response

        return wrapped_get_version()

    def _get_version(self):
        """Get API version. Simple way to test valid connectivity"""
        response = self._get_version_request()
        self._raise_if_http_error(response)
        return response.text

    def _run_emulator_request(self, qasm_path: Union[Path, str], target_system: str, num_shots: int,
                              noise_percentage: Optional[int] = None, allow_transpilation: bool = False):
        @self.auth.request_handler
        def wrapped_run_emulator(url: str, headers: dict):
            """Get the response from a qasm string upload to the emulator"""
            with open(qasm_path, 'rb') as file:
                files = {'file': file}
                data = {'shots': num_shots,
                        'target_system': target_system,
                        'noise_percentage': noise_percentage,
                        'allow_transpilation': allow_transpilation}
                response = requests.post(f"{url}/api/v1/emulator",
                                         headers=headers,
                                         data=data,
                                         files=files)
            return response

        return wrapped_run_emulator()

    def _run_get_emulator_target_system_request(self):
        @self.auth.request_handler
        def wrapped_run_emulator(url: str, headers: dict):
            """Get the response from a qasm string upload to the emulator"""
            response = requests.get(f"{url}/api/v1/emulator/backend",
                                    headers=headers)
            return response
        return wrapped_run_emulator()

    def _run_get_emulator_coupling_map_request(self, target_system):
        @self.auth.request_handler
        def wrapped_run_emulator(url: str, headers: dict):
            """Get the response from a qasm string upload to the emulator"""
            response = requests.get(f"{url}/api/v1/emulator/coupling/{target_system}",
                                    headers=headers)
            return response
        return wrapped_run_emulator()

    @staticmethod
    def _format_emulator_result(run_result_dict: dict):
        return run_result_dict  # trivial formatting in that case

    def _run_queue_request(self, start_index, end_index):
        @self.auth.request_handler
        def wrapped_run_queue(url: str, headers: dict, first: int = start_index, last: int = end_index):
            """Get the response from a request for user experiments queue"""
            response = requests.get(f"{url}/api/v1/user-experiments?first={first}&last={last}",
                                    headers=headers)
            return response

        return wrapped_run_queue()

    def _get_experiments(self, start_index: int, end_index: int) -> dict:
        """Get user queue of experiments"""
        response = self._run_queue_request(start_index=start_index, end_index=end_index)
        self._raise_if_http_error(response)
        try:
            queue = self._format_experiment_list(response.json())
        except JSONDecodeError:
            raise IndexError('No corresponding experiments could be found')
        pprint.pprint(queue)
        return queue

    @staticmethod
    def format_timestamps(experiments: dict):
        for experiment in experiments:
            experiment['timestamp'] = datetime.fromtimestamp(experiment['timestamp']).strftime('%c')
        return experiments

    def _format_experiment_list(self, queue: dict):
        output = []
        for index in queue.keys():
            output.append(queue[index])
        output = self.format_timestamps(output)
        return output

    @staticmethod
    def _check_file_extension(file: str):
        if '.qasm' not in file:
            raise NotImplementedError('Currently only .qasm files are accepted.')


class Experiment(Client):
    """
    Represents user experiments. Used to create remote jobs and retrieve results or statuses of existing ones
    """
    def __init__(self, circuit_filepath: Union[Path, str] = None, uid: str = None, get_results=True):
        super().__init__()
        assert circuit_filepath is not None or uid is not None, \
            'Please specify either the id of a previous experiment or a custom circuit file to create a new one'
        self.response = None
        self.results = None
        self.circuit_filepath = circuit_filepath
        self.uid = uid
        self.status = None
        self.qasm = None
        if uid is not None and get_results:
            self._get_results()

    def run(self):
        """Run an experiment"""
        assert self.uid is None, 'Experiment has already been triggered'
        self._run_experiment()
        print('\n Sending request to server')
        print('\n Experiment queued')

    def _run_experiment(self):
        response = self._post_run_request()
        self._raise_if_http_error(response)
        self.uid = response.text

    def _process_circuit_file(self):
        """Process a circuit file for a custom circuit upload"""
        with open(self.circuit_filepath, 'rb') as f:
            file_data = BytesIO(f.read())
        file_data.seek(0)
        return FileStorage(file_data, filename='file')

    def _post_run_request(self, chassis: str = CHASSIS_NAME, routine: str = 'custom', routine_args: dict = None):
        """Get experiment status response"""
        payload = {'chassis': chassis, 'routine': routine, 'routine_args': routine_args}
        json_data = json.dumps(payload)

        @self.auth.request_handler
        def wrapped_run_request(url: str, headers: dict):
            """Get the status of an experiment"""
            headers['json_data'] = json_data
            with open(self.circuit_filepath, 'rb') as file:
                files = {'file': file}
                response = requests.post(f'{url}/api/v1/run-routine',
                                         files=files,
                                         headers=headers)
            return response

        return wrapped_run_request()

    def get_status(self, get_results: bool = True):
        """Return the status of an experiment and populate the status attribute"""
        if self.status != 'COMPLETE':
            self.status = self._get_status()
            print(f'Status is: {self.status}')
            if self.status == 'COMPLETE' and get_results:
                print('Retrieving results')
                self._get_results()
            return
        print(f'Status is: {self.status}')
        return

    def _get_status_request(self, experiment_id: str):
        """Get experiment status response"""

        @self.auth.request_handler
        def wrapped_get_status(url: str, headers: dict):
            """Get the status of an experiment"""
            response = requests.get(f"{url}/api/v1/status/{experiment_id}",
                                    headers=headers)
            return response

        return wrapped_get_status()

    def _get_status(self):
        """Get experiment status"""
        assert self.uid is not None, 'Experiment has not yet been initialised. No valid ID found'
        response = self._get_status_request(self.uid)
        self._raise_if_http_error(response)
        self.response = response
        data = response.json()
        return data['status']

    def get_register(self):
        """Return the quantum register"""
        response = self._get_register()
        self._raise_if_http_error(response)
        return self.deserialise_array(response.json().get('results'))

    def _get_results_request(self):
        """Get experiment status response"""

        @self.auth.request_handler
        def wrapped_get_results(url: str, headers: dict):
            """Get the status of an experiment"""
            response = self._get_results_call(url, headers)
            return response

        return wrapped_get_results()

    def _get_register(self):
        @self.auth.request_handler
        def wrapped_get_register(url: str, headers: dict):
            """Get the status of an experiment"""
            headers['full_results'] = 'True'
            response = self._get_results_call(url, headers)
            return response

        return wrapped_get_register()

    def _get_results_call(self, url, headers):
        return requests.get(f"{url}/api/v1/result-distribution/{self.uid}",
                            headers=headers)

    def _get_results(self):
        """Retrieve routine results"""
        response = self._get_results_request()
        self._raise_if_http_error(response)
        self.results = response.json().get('results')
        self.qasm = response.json().get('qasm')

    def show_results(self):
        """Print the state populations of the measurement"""
        pprint.pprint(self.results)

    def plot(self):
        """Plot the routine results as a histogram"""
        assert self.results is not None, 'Results must be retrieved before they can be plotted'
        plot(self.results)

    def cancel(self):
        """Cancel an experiment that has not yet started running, or completed"""
        response = self._post_cancel_request()
        self._raise_if_http_error(response)
        if response.ok:
            print('Job successfully cancelled')

    def _post_cancel_request(self):
        """Get experiment status response"""
        @self.auth.request_handler
        def wrapped_post_cancel(url: str, headers: dict):
            """Get the status of an experiment"""
            response = requests.post(f"{url}/api/v1/cancel/{self.uid}",
                                     headers=headers)
            return response
        return wrapped_post_cancel()

    def deserialise_array(self, data: str) -> Optional[npt.NDArray]:
        """Deserialise an array and construct a response in case of failure"""
        try:
            array_bytes = base64.b64decode(data)
            np_bytes = BytesIO(array_bytes)
            return load(np_bytes, allow_pickle=True)
        except (TypeError, UnpicklingError, EOFError, ValueError) as ex:
            print('Results file parsing failed with: ')
            print(ex)
            return None



from functools import wraps
from time import sleep
from pathlib import Path
from typing import Union

import requests


class Authorisation:
    """Authenticating with auth server"""

    def __init__(self, url=None):
        self.access_token = None
        self.refresh_token = None
        self.header = None
        self.urls = url
        if url is None:
            self.url = 'https://api.seeqc.cloud'

    def authenticate(self, credentials: dict = None, refresh_token_path: Union[Path, str] = None):
        """Exchange credentials for tokens"""
        response = self._auth_request_from_cred_or_token(credentials, refresh_token_path)
        if response.status_code == 200:
            tokens = response.json()
            self.access_token = tokens.get('access')
            self.refresh_token = tokens.get('refresh')
            self.construct_header()
            print('\nAuthentication successful')
        else:
            print('Invalid credentials')
        return response.status_code

    def refresh(self) -> bool:
        """Refresh the access token and retry on failure"""
        retries = 3
        for _ in range(retries):
            is_refreshed = self.is_valid_refresh_attempt()
            if is_refreshed:
                return True
            else:
                sleep(0.5)
        print('Request could not be made')
        return False

    def is_valid_refresh_attempt(self) -> bool:
        """Refresh access token and return boolean indication success"""
        auth_url = self.url
        response = requests.post(auth_url + '/api/v1/refresh', data={'refresh': self.refresh_token})
        if response.status_code == 200:
            tokens = response.json()
            self.access_token = tokens.get('access')
            self.refresh_token = tokens.get('refresh')
            self.construct_header()
            return True
        return False

    def update_urls(self, url):
        self.url = url

    def construct_header(self):
        """Construct header in format for api gateway"""
        self.header = {"Authorization": f"Bearer {self.access_token}"}

    def request_handler(self, request_function: callable) -> callable:
        """Decorator function for API calls. Discovers API url when not specified and constructs auth header"""
        @wraps(request_function)
        def decorated_fn(*args, **kwargs):
            if 'url' not in kwargs:
                kwargs['url'] = self.url
            response = request_function(*args, **kwargs, headers=self.header)
            response.close()
            if response.status_code == 401:
                if self.refresh():
                    response = request_function(*args, **kwargs, headers=self.header)
                else:
                    print('Your authenticated session has expired and this client is unable to reauthenticate. '
                          'Please try reinstantiating the client.')
            return response
        return decorated_fn

    def gen_refresh_token(self, credentials: dict, refresh_token_path: Union[Path, str]):
        """Function used to generate a user token that can later be used for authentication"""
        auth_url = self.url
        response = requests.post(auth_url+'/api/v1/authenticate', data=credentials)
        if response.status_code == 200:
            print('\nAuthentication successful, retrieving refresh token')
            tokens = response.json()
            refresh_token = tokens.get('refresh')
            with open(refresh_token_path, 'w') as f:
                f.write(refresh_token)
        else:
            raise PermissionError('Invalid credentials')

    def _auth_request_from_cred_or_token(self, credentials: dict = None,
                                                refresh_token_path: Union[Path, str] = None):
        """Create and submit a post request to authenticate using either credentials or a token"""
        if (credentials is None) and (refresh_token_path is None):
            raise ValueError("Need to specify an authentication method: either credentials or refresh token")
        if (credentials is not None) and (refresh_token_path is not None):
            raise ValueError("Can't specify both credentials and a refresh token during authentication")
        auth_url = self.url
        response = None
        if credentials is not None:
            response = requests.post(auth_url+'/api/v1/authenticate', data=credentials)
        else:
            assert refresh_token_path is not None
            with open(refresh_token_path, 'r') as f:
                refresh_token_str = f.read()
                refresh_token_data = {'refresh': refresh_token_str}
                response = requests.post(auth_url+'/api/v1/refresh', data=refresh_token_data)
        return response

    def send_password_reset_email(self, email: str):
        response = requests.post(self.url+'/api/v1/reset-password', data={'email': email})
        print(response.text)
        if response.ok:
            'Password reset email has been sent. Try checking your junk email folder if it is not in your inbox.'
        else:
            'These was a problem fulfilling your request. '
            'Either try again, ensuring your email was spelled correctly or contact the administrator.'



from __future__ import annotations
from .test_vars import QueueTestVars, EXPERIMENT_ID

VALID_PASSWORD = 'valid'
VALID_USERNAME = 'user'
VALID_ACCESS = 'valid_access'
VALID_REFRESH = 'valid_refresh'
VERSION = 'version_value'


class MockedAuthenticate:
    """Mock replacement for Authenticate"""
    def __init__(self, *args, **kwargs):
        self.status_code = None
        self.access_token = None
        self.refresh_token = None
        self.mocked_authenticate(*args, **kwargs)

    def json(self):
        return {
            'access': self.access_token,
            'refresh': self.refresh_token
        }
    
    def mocked_authenticate(self, *args, **kwargs):
        """authentication function to check credential validity """
        data = kwargs.get('data')
        email = data.get('email')
        password = data.get('password')
        if email != VALID_USERNAME or password != VALID_PASSWORD:
            self.status_code = 400
            return self
        self.status_code = 200
        self.access_token = VALID_ACCESS
        self.refresh_token = VALID_REFRESH
        return self


class MockedRefresh:
    """Mock class to handle token refreshing"""
    def __init__(self, *args, **kwargs):
        self.access_token = None
        self.refresh_token = None
        self.status_code = None
        self.mocked_refresh(*args, **kwargs)

    def mocked_refresh(self, *args, **kwargs):
        """Replace tokens with valid values"""
        data = kwargs.get('data')
        refresh = data.get('refresh')
        if refresh == VALID_REFRESH:
            self.access_token = VALID_ACCESS
            self.refresh_token = VALID_REFRESH
            self.status_code = 200
        else:
            self.status_code = 400

    def json(self):
        return {
            'access': self.access_token,
            'refresh': self.refresh_token
        }


class MockedVersion:
    """Mocking class to replace the get version API route"""
    def __init__(self, *args, **kwargs):
        self.status_code = None
        self.ok = False
        self.text = None
        self.mocked_version(*args, **kwargs)

    def mocked_version(self, *args, **kwargs):
        """Check tokens and return a verion string"""
        access = extract_token_from_header(kwargs)
        if access == VALID_ACCESS:
            self.status_code = 200
            self.ok = True
            return VERSION
        else:
            self.status_code = 401
            self.text = 'error'

    @staticmethod
    def close():
        pass


class MockedEmulator:
    """Mocking class to replace the emulator"""
    def __init__(self, *args, **kwargs):
        self.status_code = None
        self.text = None
        self.allowed_extensions = {'txt', 'qasm'}
        self.ok = False
        self.text = None
        self.mocked_emulator(*args, **kwargs)

    def mocked_emulator(self, *args, **kwargs):
        """Check files and return a response consistent with emulator api route"""
        file = kwargs['files']['file']
        filename = file.raw.name
        if self.allowed_file(filename):
            self.status_code = 200
            self.ok = True
        else:
            self.status_code = 400
            self.text = 'error'
        self.text = '[[0],[0],[0],[0],[0],[0],[0],[0],[0],[0]]\n'

    def allowed_file(self, filename):
        """Check file is as expected"""
        return '.' in filename and \
            filename.rsplit('.', 1)[1].lower() in self.allowed_extensions

    @staticmethod
    def close():
        pass


class MockedQueue:
    """Replacing the status-queue API route"""
    def __init__(self, *args, **kwargs):
        self.status_code = None
        self.queue = None
        self.ok = False
        self.text = None
        self.mocked_queue(*args, **kwargs)

    @staticmethod
    def parse_path(path: str):
        """Get path arguments from a url"""
        short_url = path.split("?", 1)[1]
        split_args = short_url.split('&')
        arguments = {}
        for item in split_args:
            key, val = item.split('=')
            arguments[key] = int(val)
        return arguments

    def mocked_queue(self, *args, **kwargs):
        """Generate a mocked queue of experiments"""
        arguments = self.parse_path(args[0])
        full_queue = {
            0: {'experiment_id': QueueTestVars.experiment_id0, 'status': 'PENDING', 'timestamp': 1686057785.409567},
            1: {'experiment_id': QueueTestVars.experiment_id1, 'status': 'COMPLETED', 'timestamp': 1686057785.441656},
            2: {'experiment_id': QueueTestVars.experiment_id2, 'status': 'COMPLETED', 'timestamp': 1686057785.456754},
            3: {'experiment_id': QueueTestVars.experiment_id3, 'status': 'COMPLETED', 'timestamp': 1686057786.456754},
            4: {'experiment_id': QueueTestVars.experiment_id4, 'status': 'COMPLETED', 'timestamp': 1686057787.456754}
        }
        self.remove_out_of_range_indices(full_queue, arguments)
        self.queue = full_queue
        self.status_code = 200
        self.ok = True

    def json(self):
        return self.queue

    @staticmethod
    def remove_out_of_range_indices(queue: dict, arguments: dict):
        """Use path arguments to filter out of range indices from a queue"""
        bad_indices = []
        if 'first' in arguments.keys() and 'last' in arguments.keys():
            for index in queue.keys():
                if index < int(arguments['first']):
                    bad_indices.append(index)
                if index > int(arguments['last']):
                    bad_indices.append(index)
            for index in bad_indices:
                del queue[index]

    @staticmethod
    def close():
        pass


class MockedStatus:
    """Top level class for get status mocking"""
    def __init__(self, *args, **kwargs):
        self.status_code = None
        self.ok = False
        self.text = None

    @staticmethod
    def close():
        pass


class MockedPendingStatus(MockedStatus):
    """Class to return a pending status"""
    def __init__(self, *args, **kwargs):
        super().__init__()

    def get_status(self):
        self.ok = True
        return {
            'experiment_id': EXPERIMENT_ID,
            'status': 'PENDING',
            'timestamp': 1686143323.307925
        }

    def json(self):
        return self.get_status()


class MockedCompleteStatus(MockedStatus):
    """Class to return a complete status"""
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.json = self.get_status()

    def get_status(self):
        self.ok = True
        return {
            'experiment_id': EXPERIMENT_ID,
            'status': 'Complete',
            'timestamp': 1686143323.307925
        }


class MockedResults:
    """Mocking class to return a results json object"""
    def __init__(self):
        self.json = {'0': 0.75, '1': 0.25}
        self.status_code = None

    def close(self):
        pass


def extract_token_from_header(headers: dict[dict[str]]):
    """Get auth token from an HTTP header"""
    bearer = headers.get('headers').get('Authorization')
    return bearer.replace('Bearer ', '')



import unittest
from unittest.mock import patch

from seeqc_client import Client, Experiment
from .mocked_routes import VALID_PASSWORD, VALID_USERNAME, VALID_ACCESS, VALID_REFRESH, MockedAuthenticate, \
    MockedRefresh, MockedVersion, MockedEmulator, MockedQueue, MockedPendingStatus
from seeqc_client.tests.test_vars import QueueTestVars, EXPERIMENT_ID


class BaseCase(unittest.TestCase):
    """
    CLI Tests
    """
    def setUp(self):
        """Setup tests"""
        self.credentials = {'email': VALID_USERNAME, 'password': VALID_PASSWORD}
        self.client = Client()

    def tearDown(self) -> None:
        self.client.auth.access_token = None
        self.client.auth.refresh_token = None
    
    @patch('requests.post', MockedAuthenticate)
    def test_valid_authenticate(self):
        """Test that a client successfully authenticates given the correct credentials"""
        self.client.auth.authenticate(self.credentials)
        self.assertEqual(self.client.auth.access_token, VALID_ACCESS)
        self.assertEqual(self.client.auth.refresh_token, VALID_REFRESH)

    @patch('requests.post', MockedAuthenticate)
    def test_invalid_authenticate(self):
        """Test a client fails to authenticate with bad credentials"""
        self.client.auth.authenticate({'email': 'invalid', 'password': 'invalid'})
        self.assertEqual(self.client.auth.access_token, None)
        self.assertEqual(self.client.auth.refresh_token, None)

    @patch('requests.post', MockedRefresh)
    def test_valid_refresh(self):
        """Test that tokens can be refreshed given a valid refresh token"""
        self.client.auth.refresh_token = VALID_REFRESH
        self.client.auth.refresh()
        self.assertEqual(self.client.auth.access_token, VALID_ACCESS)

    @patch('requests.post', MockedRefresh)
    def test_invalid_refresh(self):
        """Test refresh fails with an invalid token"""
        self.client.auth.refresh_token = 'invalid'
        self.client.auth.refresh()
        self.assertEqual(self.client.auth.access_token, None)

    def test_get_version_valid(self):
        """Test that a successful request can be made to the resource server"""
        with patch('requests.post', MockedAuthenticate):
            self.client.auth.authenticate(self.credentials)
        with patch('requests.get', MockedVersion):
            response = self.client._get_version_request()
        self.assertEqual(response.status_code, 200)

    @patch('requests.post', MockedRefresh)
    def test_get_version_invalid_access_and_refresh(self):
        """Test that an unsuccessful request can be made to the resource server"""
        self.client.auth.access_token = 'invalid'
        self.client.auth.construct_header()
        with patch('requests.get', MockedVersion):
            response = self.client._get_version_request()
        self.assertEqual(response.status_code, 401)

    def test_post_good_emulator_request(self):
        """Test that a successful request can be made to an emulator backend"""
        with patch('requests.post', MockedEmulator):
            response = self.client._run_emulator_request('seeqc_client/tests/test.qasm', '', 10)
        self.assertEqual(response.status_code, 200)

    def test_post_bad_emulator_request(self):
        """Test that a successful request can be made to an emulator backend"""
        with patch('requests.post', MockedEmulator):
            response = self.client._run_emulator_request('seeqc_client/tests/bad_extension.qasm2', '', 10)
            self.assertEqual(response.status_code, 400)

    def test_get_experiments(self):
        """Check experiment queue get"""
        with patch('requests.get', MockedQueue):
            data = self.client.get_experiments()
        self.assertEqual(data[0]['experiment_id'], QueueTestVars.experiment_id0)
        self.assertEqual(len(data), 5)

    def test_get_experiments_private(self):
        """Check experiment queue get private method"""
        with patch('requests.get', MockedQueue):
            response = self.client._run_queue_request(start_index=0, end_index=20)
        self.assertEqual(response.status_code, 200)
        data = response.queue
        self.assertEqual(data[0]['experiment_id'], QueueTestVars.experiment_id0)
        self.assertEqual(len(data.keys()), 5)

    def test_get_experiments_short(self):
        """Check experiment queue with shortened queue"""
        with patch('requests.get', MockedQueue):
            response = self.client._run_queue_request(start_index=1, end_index=3)
        self.assertEqual(response.status_code, 200)
        data = response.queue
        self.assertEqual(data[1]['experiment_id'], QueueTestVars.experiment_id1)
        self.assertFalse(QueueTestVars.experiment_id4 in data.keys())
        self.assertFalse(QueueTestVars.experiment_id0 in data.keys())
        self.assertEqual(len(data.keys()), 3)

    def test_get_experiments_too_long(self):
        """Test experiment queue fails when too many experiments requested"""
        with self.assertRaises(AssertionError):
            self.client.get_experiments(start_index=0, end_index=Client().max_queue_length+1)

    def test_get_experiments_bad_range(self):
        """Test experiment queue fails when indices are given in wrong order"""
        with self.assertRaises(AssertionError):
            self.client.get_experiments(start_index=10, end_index=0)

    def test_pending_get_status(self):
        """Test that when an experiment is found to be pending that the results are not fetched"""
        experiment = Experiment(uid=EXPERIMENT_ID, get_results=False)
        with patch('requests.get', MockedPendingStatus), \
             patch('seeqc_client.Experiment._get_status') as mock_results:
            status = experiment._get_status()
            mock_results.assert_called_once()
            self.assertIsNotNone(status)

    def test_get_results(self):
        """Test that a results request is made as expected"""
        experiment = Experiment(uid=EXPERIMENT_ID, get_results=False)
        with patch('seeqc_client.Experiment._get_results_request') as mocked:
            experiment._get_results()
            mocked.assert_called_once()

    @patch('getpass.getpass')
    @patch('seeqc_client.Client._get_username')
    def test_initialise(self, mock_user, mock_getpass):
        """Test that a user can initialise the client"""
        mock_getpass.return_value = VALID_PASSWORD
        mock_user.return_value = VALID_USERNAME
        with patch('seeqc_client.authorisation.Authorisation.authenticate') as mocked:
            self.client.initialise()
            mocked.assert_called_once_with(credentials=self.credentials)
        mock_user.reset_mock()
        mock_getpass.reset_mock()

    def test_cancel(self):
        experiment = Experiment(uid=EXPERIMENT_ID, get_results=False)
        with patch('seeqc_client.Experiment._post_cancel_request') as mock_cancel:
            experiment.cancel()
            mock_cancel.assert_called_once()

    def test_format_result(self):
        """Get experiment status response"""
        result = { "measurement_outcomes": { "0": 2, "1": 0 } }
        formatted = self.client._format_emulator_result(result)
        self.assertEqual(formatted, result)



from dataclasses import dataclass

@dataclass
class QueueTestVars:
    experiment_id0 = '0'
    experiment_id1 = '1'
    experiment_id2 = '2'
    experiment_id3 = '3'
    experiment_id4 = '4'

EXPERIMENT_ID = 'b1e5d259-f0fd-4ccb-b0f2-455394b84b4d'



import os

import matplotlib as mpl
import matplotlib.image as image
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.offsetbox import OffsetImage, AnnotationBbox


def _get_logo_path() -> str:
    """return path of Seeqc logo"""
    import seeqc_client
    path = os.path.dirname(seeqc_client.__file__)
    return path + r'\black_logo.png'


def plot(data: dict):
    """plot results histogram"""
    mpl.rcParams["font.family"] = "Consolas"
    file = _get_logo_path()
    seeqc = image.imread(file)
    names = list(data.keys())
    values = list(data.values())
    fig = plt.bar(range(len(data)), values, tick_label=names, color='black')
    ax = plt.gca()
    ax.set_facecolor("#e4e4e3")
    ax.set_ylim([0, 1])
    ax.grid(axis='y', linestyle=(0, (3, 5)))
    ax.set_axisbelow(True)
    ax.set_ylabel('Populations')
    plt.yticks(np.arange(0, 1.1, 0.1))
    imagebox = OffsetImage(seeqc, zoom=0.1, alpha=0.8)
    position = ((len(data)-1)*0.5, 0.5)  # (x, y) position, with x in the center of bars and y at the middle of the y-axis range
    ab = AnnotationBbox(imagebox, position, frameon=False, xycoords='data')
    ax.add_artist(ab)
    plt.show()
